#include<linux/init.h>          // Needed for basic initialization.
#include<linux/module.h>        // Needed for every module
#include<linux/fs.h>            // The file operations defined here.
#include<linux/cdev.h>          // The cdev structure is defined in it.
#include<linux/slab.h>          // Needed as kmalloc() is declared in it.
#include<linux/device.h>        // The device creation functions are defined here.
#include<linux/interrupt.h>     // Needed to register and handle the interrupt.
#include<linux/delay.h>         // To use delay function ssleep().
#include<linux/uaccess.h>       // Needed by segment descriptor.
#include<linux/kernel.h>        // Allow to use printk function.
#include<linux/string.h>        // needed for string operations. 

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


MODULE_LICENSE("GPL");                                       		 // This is a GPL-licesed module.        
MODULE_DESCRIPTION("Interrupt handling in the device driver.\n");        // Macro gives short description of a module.

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define IRQ_NUM 2                  // The IRQ line number.
static dev_t device_num;              // Global variable that holds the device number.
static struct class *cl;          // structure creating the class for a device.  

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* cdev structure for a character device .*/

static struct cdev my_cdev;      

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The prototypes of the file operations such as open, read, write and release that implemented in this driver */

static int mydevice_open (struct inode *inod, struct file *filp);
static int mydevice_release (struct inode *inod, struct file *filp);

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The file_operation structure of a device.*/

static struct file_operations fops=
{
.open= mydevice_open,
.release= mydevice_release,

};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The interrupt handler.*/

irqreturn_t irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
		
		printk(KERN_INFO "my_driver : Inside the interrupt handler function.\n");
		return IRQ_HANDLED;

}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The Initialization function */

static int init_function(void)
{
	printk(KERN_INFO "my_driver : Inside the initialization function.\n");
	
	if(alloc_chrdev_region(&device_num, 0, 1, "my_device") < 0)
	{	
		printk(KERN_INFO "my_driver : the device number allocation fails.\n");
		return -1;
	}
	printk(KERN_INFO "my_driver  : the device number allocation done successfully.\n");	
	printk(KERN_INFO "my_driver  : major and minor numbers of my_device are : %d, %d\n", MAJOR(device_num),  MINOR(device_num));


	if((cl = class_create(THIS_MODULE,"my_class")) == NULL)
	{
		
		printk(KERN_INFO "my_driver  : The class creation failed.\n");		
		unregister_chrdev_region(device_num,1);
		return -1;
	}
	else
	{
		printk(KERN_INFO "my_driver  : The class is created successfully.\n");

	}
	
	if(device_create(cl,NULL,device_num,NULL,"new_device") == NULL)
	{
		printk(KERN_INFO "my_driver  : The device node creation fails here.\n");		
		class_destroy(cl);
		unregister_chrdev_region(device_num,1);
		return -1;
	}
	else 
        {
		printk(KERN_INFO "my_driver  : The device node is created successfully.\n");
	
	}
		
	cdev_init(&my_cdev,&fops);
	if(cdev_add(&my_cdev,device_num,1) == -1)
	{
		device_destroy(cl,device_num);
		class_destroy(cl);
		unregister_chrdev_region(device_num,1);
		return -1;
	}
	else
	{
		printk(KERN_INFO "my_driver  : The cdev_structure initialization done successfully.\n");
			
	}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* 
int request_irq(IRQ_NO, (irq_handler_t) irq_handler, flag, "device_interrupt_name",(void *)(irq_handler)) - This function is used to register the interrupt handler.

*/

	
	if (request_irq(IRQ_NUM, (irq_handler_t) irq_handler, IRQF_SHARED, "my_interrupt", (void *)irq_handler))
        {
        	printk(KERN_ALERT "my_driver  : Failed to register interrupt handler of irq 2.\n");
        	return -EIO;
    	}
	printk(KERN_ALERT "my_driver  : Interrupt handler of irq 2 registered successfully.\n");
	return(0); 

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* the open function of a device */

static int mydevice_open (struct inode *inod, struct file *filp)
{
	  
	printk(KERN_INFO "my_driver  : The device opened successfully.\n");

	printk(KERN_INFO "my_driver :  The interrupt hits here.\n");	

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*  The IRQ0 is mapped to vector using the macro - #define IRQ0_VECTOR (FIRST_EXTERNAL_VECTOR + 0x10)
    To raise an interrupt IRQ 2, programmatically we have to add 2 to a vector of IRQ0.
    0x20 + 0x10 + 02 = 50 (in Decimal)
*/
 	asm("int $50");
	printk(KERN_INFO "my_driver : Exit from the device open function.\n");

	return 0;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The close function of a device */

static int mydevice_release(struct inode *inod, struct file *filp)
{
	printk(KERN_INFO "my_driver : The device is closed successfully.\n");
	
	return(0);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The cleanup function */

static void exit_function(void)
{
	printk(KERN_INFO "my_driver : Inside the exit function of a Driver.\n");

	cdev_del(&my_cdev);
	printk(KERN_INFO "my_driver : The cdev structure removed from the system.\n");
	
	device_destroy(cl,device_num);
	printk(KERN_INFO "my_driver : The device removed from the /dev directory and the system.\n");

	class_destroy(cl);
	printk(KERN_INFO "my_driver : The class of the device removed from the system.\n");
	
	unregister_chrdev_region(device_num,1);
	printk(KERN_INFO "my_driver : the device unregistration done successfully...\n");

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*  
using free_irq(IRQ_NUM,(void *)irq_handler) - This function is used to remove the interrupt handler from the kernel. 
*/

	 free_irq(IRQ_NUM, (void *)irq_handler);
	 printk(KERN_INFO "my_driver : The interrupt handler removed from the kernel.\n");
	
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The init and exit functions of the driver are registered here. */

module_init(init_function);
module_exit(exit_function);









